#
# -*- Mode:Perl; -*-
#
# avoidLevel controls the level of routine that should be avoided (setting this
# at 3 or lower suppresses some warnings).
$avoidLevel = 3;

# FilesUsingFunc let us summarize the problems found at the end
# The key is "filename;function", we can then sort on the filenames.
# The value of the entry is the number of times that the function appeared
# in that files
%FilesUsingFunc = ();
#
# Subroutines that we want to avoid.  The associated value is the level
# at which we want to avoid them:
#   1: Always wrong 
#   2: Discouraged, better substitutes exist
#   3: Discouraged but sometimes ok during development
# $avoidLevel is set to the level permissible (3 by default)
%avoidsubs = ( "fprintf" => 2, "fputs" => 1, "puts" => 1, "printf" => 2, 
               "malloc" => 1, "calloc" => 1, "free" => 1, "strdup" => 1, 
	       "realloc" => 1,
	       "assert" => 3, "perror" => 3, "getenv" => 3 );

# avoidSubsReasons gives the reasons for avoiding each of these 
$AvoidPrint = "\
For internationalization, messages must be indirected through a routine such as
gettext.  Marking the print routines makes it possible to catch oversights in
writing messages for the user.  Alternate routines are provided to providing
printed output for debugging or for the developer.";

$AvoidMalloc = "\
Dynamic memory allocation often causes problems, though the creation of both 
storage leaks and memory overwrites.  By using the MPIU_xxx names for memory 
allocation, it is easy to choose between the standard malloc/free/etc. versions
and special versions that provide detailed memory correctness diagnostics, 
without requiring any special third-party software (which makes these 
particularly appropriate when the problem occurs only when running on hundreds 
of nodes).";

$IntPointerMalloc = "\
Routines that return pointers must be declared as such to vavoid failures on 
platforms where pointers and integers have different sizes, such as Alphas 
and IA64s.";

#
# Added extra information to strdup because of failures in ROMIO code that
# uses strdup without declaring it (failed on IA64)
%avoidSubsReasons = ( "fprintf" => $AvoidPrint,
		      "fputs" => $AvoidPrint,
		      "puts" => $AvoidPrint,
		      "printf" => $AvoidPrint,
		      "malloc" => $AvoidMalloc,
		      "calloc" => $AvoidMalloc,
		      "free" => $AvoidMalloc,
		      "strdup" => $AvoidMalloc . "\n" . $IntPointerMalloc,
		      "realloc" => $AvoidMalloc,
		      "assert" => "Assert is a handy way to check code while developing, but in released code, these cases should be handled with more attention to the message the the user will receive.",
		      "perror" => "For internationalization reasons, messages should use the internationalization support instead of perror.", 
		      "getenv" => "In parallel program environments, there is no guarantee that environment variables will be available to all processes." ); 

# avoidSubsSeen keeps track of which of these routines were seen (including
# the count).
%avoidSubsSeen = ();

#
# Guarded subs maps the routines being guarded to the guard definition
# (we may eventually use colon separated values if multiple definitions are
# allowed)
%guardedsubs = ( "alloca" => "HAVE_ALLOCA", 
		 "closedir" => "HAVE_CLOSEDIR", 
		 "fnmatch" => "HAVE_FNMATCH", 
		 "getloadavg" => "HAVE_GETLOADAVG", 
		 "getmntent" => "HAVE_GETMNTENT", 
		 "getpgrp" => "HAVE_GETPGRP", 
		 "memcmp" => "HAVE_MEMCMP", 
		 "mmap" => "HAVE_MMAP",    # Not really good enough
		 "setpgrp" => "HAVE_SETPGRP", 
		 # "setvbuf" => "HAVE_SETVBUF", 
		 "strcoll" => "HAVE_STRCOLL", 
		 "strftime" => "HAVE_STRFTIME",
		 "utime" => "HAVE_UTIME", 
		 "vfork" => "HAVE_VFORK", 
		 "vprintf" => "HAVE_VPRINTF", 
		 "wait3" => "HAVE_WAIT3" );

# setvbuf is apparently part of C89, so I've removed it from the
# guarded functions.  But see SETVBUF_REVERSED and
# AC_FUNC_SETVBUF_REVERSED

# Removed strcmp because it is ok if one string is guaranteed to 
# have a terminating null.  This is still touchy, but it isn't the risk
# posed by the copy or cat routines, since it doesn't write to any memory
# location.

#
# FIXME: bettersubs and avoidsubs should be unified.  Bettersubs is really
# a low-level avoid.  In addition, strcpy can often be considered an avoidsub 
# rather than just a sub with a better version (never, ever use the string 
# routines that don't take the length of the destination buffer).

# strerrror is ANSI C, so we allow it without warning until we have a
# system that fails to build with strerror
%bettersubs = ( "assert" => "MPIU_Assert or MPIU_Assertp",
		"strcpy" => "MPIU_Strncpy",
	        "strncpy" => "MPIU_Strncpy",
#		"strcmp" => "strncmp",
		"strcat" => "MPIU_Strnapp",
		"sprintf" => "MPIU_Snprintf", 
		"snprintf" => "MPIU_Snprintf",
#		"strerror" => "MPIU_Strerror",
		);

$BetterString = "\
The standard string routines are risky because they provide no protection for 
input strings that are too long for the output strings.  Some of the \"n\" 
versions (e.g., strncat) are also risky because the length parameter is not 
the length of the buffer, and hence can be computed incorrectly.";
%betterSubsReasons = (
		      "assert" => "\
Assert is a handy way to include sanity checks in a code during development, 
but it does not generate good error messages should the assert fail for a user.
Instead, it is better to use customized assert routines that will provide a 
more user-centric error message.",
		      "strcpy" => $BetterString,
		      "strncpy" => $BetterString,
		      "strcmp"  => $BetterString,
		      "strcat"  => $BetterString,

		      "sprintf" => "\
Sprintf, like the standard string routines, is risky because the input 
parameters can create an output string that is too long for the provided 
output buffer, with no way to detect this and indicate an error.",
		      "snprintf" => "\
This routine is not portable; not all environments provide it.  More 
specifically, snprintf was not part of the 1990 ANSI/IOS C standard though 
it is part of C99.  Further, while some enviroments provide an snprintf entry 
point, they implement it by calling sprintf directly with no checks on the 
provided buffer, making snprintf no safer than sprintf.  Also, C99 and 
Single Unix Specification give different return values for snprintf.  With 
all of this, it isn't even clear what routine you are using when you use 
snprintf.",
		      "strerror" => "\
This routine is not portable to older systems (it was added in C89); not all 
environments provide it.  Applications may use it but should consider guarding 
it with a HAVE_STRERROR test."
		      );
# --------------------------------------------------------------------------
%AllowedRoutines = ();
# guardedsubsGuard is used to keep track of HAVE_routinename tests
%guardedsubsGuard = ();

# These provide for modifications to the lists of allowed routines, 
# scoped by directory (just this dir or a directory tree)
@gAllowedRoutinesForDirStack = ();
@gAllowedRoutinesForDirTreeStack = ();

sub ClearAvoidFuncs {
    %AllowedRoutines = ();
    %guardedsubsGuard = ();

    for (my $i=$#gAllowedRoutinesForDirStack; $i>=0; $i--) {
	my $elm = $gAllowedRoutinesForDirStack[$i];
	my ($depth,$varname,$action) = split(/:/,$elm);
	
	foreach my $key (keys(%$varname)) {
	    $AllowedRoutines{$key} = $$varname{$key};
	}
	if ($action eq "replace") { last; }
    }

    for (my $i=$#gAllowedRoutinesForDirTreeStack; $i>=0; $i--) {
	my $elm = $gAllowedRoutinesForDirTreeStack[$i];
	my ($depth,$varname,$action) = split(/:/,$elm);
	
	foreach my $key (keys(%$varname)) {
	    $AllowedRoutines{$key} = $$varname{$key};
	}
	if ($action eq "replace") { last; }
    }

}
sub ClearDirAvoidFuncs {
    if ($#gAllowedRoutinesForDirStack >= 0) {
	my $elm = $gAllowedRoutinesForDirStack[$#gAllowedRoutinesForDirStack];
	my ($depth,$varname,$action) = split(/:/,$elm);
	if ($depth == $gDirDepth) {
	    $#gAllowedRoutinesForDirStack --;
	}
    }

}
sub ClearDirTreeAvoidFuncs {
    if ($#gAllowedRoutinesForDirTreeStack >= 0) {
	my $elm = $gAllowedRoutinesForDirTreeStack[$#gAllowedRoutinesForDirTreeStack];
	my ($depth,$varname,$action) = split(/:/,$elm);
	if ($depth == $gDirDepth) {
	    $#gAllowedRoutinesForDirTreeStack --;
	}
    }
}

sub CheckLineAvoidFuncs {
    my ($FD, $curline, $filename) =@_;

    # Only apply to C and C++ files
    if ($gLanguage ne "C" && $gLanguage ne "CXX") { return $curline; }


    foreach $subname (keys(%avoidsubs)) {
	if (/(?<!\w)$subname\s*\(/ && !defined($AllowedRoutines{$subname})) {
	    # Check for level
	    $routineLevel = $avoidsubs{$subname};
	    if ($routineLevel >= $avoidLevel) { next; }
	    $msg = "Warning: found $subname in $filename";
	    $avoidSubsSeen{$subname} = $avoidSubsSeen{$subname} + 1;
	    if (defined($warningmsgs{$msg})) { 
		$warningmsgs{$msg} = $warningmsgs{$msg} + 1;
	    }
	    else {
		$warningmsgs{$msg} = 1;
	    }
	    if (!defined($FilesUsingFunc{"$filename;$subname"})) {
		$FilesUsingFunc{"$filename;$subname"} = 0; }
	    $FilesUsingFunc{"$filename;$subname"} ++;
	}
    }

    # The guarded subs are ones that GNU autoconf is worried about.
    # This code should look for some sort of configure-style CPP value
    # and allow the ones that are properly guarded.
    # The mysterious (?<!\w) is a "negative look-behind match".  That is,
    # do *not* match a single "word" (alphanumeric+_) character.  This
    # keeps us from matching MPIU_vprintf when vprintf is in the list
    # of test routines and MPIU_vprintf is in the current line.
    
    # Note that we make use of the definitions seen to check
    # whether we should complain about the routine

    foreach $subname (keys(%guardedsubs)) {
	if (/(?<!\w)$subname\s*\(/ && !defined($AllowedRoutines{$subname})) {
	    if (!defined($defsUsed{$guardedsubs{$subname}})) {
		&PrintCurrentDirectory;
		print "Caution: found reference to routine $subname in $filename\n";
		if (!defined($FilesUsingFunc{"$filename;$subname"})) {
		    $FilesUsingFunc{"$filename;$subname"} = 0; }
		$FilesUsingFunc{"$filename;$subname"} ++;
	    }
	}
    }
    # See above for an explanation of (?<!\w)
    foreach $subname (@guardsubs) {
        if (/(?<!\w)$subname\s*\(/ && !defined($AllowedRoutines{$subname})) {
            $msg = "Caution: found $subname in $filename";
            if (defined($warningmsgs{$msg})) {
                $warningmsgs{$msg} = $warningmsgs{$msg} + 1;
            }
            else {
                $warningmsgs{$msg} = 1;
            }
	    if (!defined($FilesUsingFunc{"$filename;$subname"})) {
		$FilesUsingFunc{"$filename;$subname"} = 0; }
	    $FilesUsingFunc{"$filename;$subname"} ++;
        }
    }
#    foreach $subname (keys %bettersubs) {
#	if (/(?<!\w)$subname\s*\(/) {
#	    print "Caution: found reference to routine $subname in $filename\n";
#	    print "Consider using $bettersubs{$subname} instead\n";
#	}
#    }

    foreach $subname (keys %bettersubs) {
        if (/(?<!\w)$subname\s*\(/ && !defined($AllowedRoutines{$subname})) {
            $msg = "Caution: found $subname in $filename, consider $bettersubs{$subname}";
            if (defined($warningmsgs{$msg})) {
                $warningmsgs{$msg} = $warningmsgs{$msg} + 1;
            }
            else {
                $warningmsgs{$msg} = 1;
            }
	    if (defined($betterSubsSeen{$subname})) {
		$betterSubsSeen{$subname} = $betterSubsSeen{$subname} + 1;
	    }
	    else {
		$betterSubsSeen{$subname} = 1;
	    }
	    if (!defined($FilesUsingFunc{"$filename;$subname"})) {
		$FilesUsingFunc{"$filename;$subname"} = 0; }
	    $FilesUsingFunc{"$filename;$subname"} ++;
        }
    }
}

# Summarize the reason for any avoid subs
$headerPrinted = 0;
foreach $key (keys(%avoidSubsSeen)) {
    my $reason = $avoidSubsReasons{$key};
    if (!$headerPrinted && $reason ne "") {
	$headerPrinted = 1;
	print "${topHeader}Explanation for messages about routines to avoid\n";
	$topHeader = "";
    }
    print "\n$key: $reason\n";
}

sub FinalReportOnAvoidFuncs {
    # Summarize the files with functions
    my $filename = "";
    my $headerPrinted = 0;
    foreach my $key (sort(keys(%FilesUsingFunc))) {
	my ($file,$func) = split(/;/,$key);
	if ($file ne $filename) {
	    $filename = $file;
	    if (!$headerPrinted) {
		$headerPrinted = 1;
		print "Files using functions to be avoided:"
	    }
	    print "\n" . $filename;
	}
	print " " . $func;
    }
    print "\n";

    # Summarize the reason for any better subs
    $headerPrinted = 0;
    foreach my $key (keys(%betterSubsSeen)) {
	my $reason = $betterSubsReasons{$key};
	if (!$headerPrinted && $reason ne "") {
	    $headerPrinted = 1;
	    print "$topHeader\nExplanation for messages about routines with prefered versions\n";
	    $topHeader = "";
	}
	print "\n$key: $reason\n";
    }
}

sub DescribeAvoidFuncs {
    print STDOUT 
	"\tavoidfuncs - Check for functions that should be avoided, such as strcpy\n";
}

# ---------------------------------------------------------------------------
# Routines to allow customization of the names that are allowed
# ( name of hash containing names, tree or dir, add or replace ), 
# eg., 
#   %mynames = ( 'printf' => mypack );
#   &PushAllowFuncNames( "mynames", "tree", "replace" );
sub PushAllowFuncNames {
    my ($varname, $dirscope, $action ) = @_;
    my $elm;

    if ($action ne "add" && $action ne "replace") {
	print STDERR "Action in PushAllowFuncNames is $action, must be either add or replace";
	return 0;
    }
    $elm = $gDirDepth . ":" . $varname . ":" . $action;
    
    if ($dirscope eq "tree") {
	$gAllowedRoutinesForDirTreeStack[$#gAllowedRoutinesforDirTreeStack+1] = 
	    $elm
    }
    elsif ($dirscope eq "dir") {
	$gAllowedRoutinesForDirStack[$#gAllowedRoutinesforDirStack+1] = 
	    $elm
    }
    else {
	print STDERR "Invalid scope in PushAllowFuncNames; was $dirscope but must be either tree or dir\n";
	return 0;
    }
    return 1;
}

# --------------------------------------------------------------------------
# Add to known actions
$KnownActions[$#KnownActions+1] = "AvoidFuncs";

1;
