#
# -*- Mode:Perl; -*-
#
# Comment processing is special because it must apply its
# own set of tests only within a comment.
# The steps are as follows:
#    1) Read comment
#    2) Check for unallowed comment formats (e.g., // (a C++ comment) in 
#       a C file
#
sub ClearComments {
    $cxxCommentsFound = 0;
}

sub CheckLineComments {
    my ($FD, $curline, $filename) =@_;

    my $commentText;

    # First, get the entire comment.  Use the gLanguage to pick the
    # correct routine.  Return the pair (comment, remainder)
    ($commentText, $curline) = &ReadComment( $FD, $curline, $gLanguage );

    # Check for C++ style comments (Comment starts with //; we 
    # need to allow // within the comment)
    if ($checkComments && $commentText =~ /^\/\//) {
	$cxxCommentsFound ++;
	if ($cxxCommentsFound == 1 && ! $gCxxComments) {
	    &PrintCurrentDirectory;
	    print "C++ comment found in $filename:\n";
	    $gAvoidConstructionsSeen{"//"} .= "$filename;";
	    my $line = $commentText;
	    if ($commentText =~ /(.*)\n/) {
		$line = "$1\n";
	    }
	    else {
		$line = $_;
	    }
	    print $line;
	}
	$curline =~ s/\/\/.*//;
    }

    if ($commentText =~ /\S/) {
	# Apply if there are any comments on this line
	for (my $i=0; $i<=$#KnownActionsInComments; $i++) {
	    my $fcn = "CheckLine" . $KnownActionsInComments[$i];
	    print "Applying $fcn to $commentText\n" if $gDebug;
	    $commentText = &$fcn( FD, $commentText, $filename );
	}
    }

    return $curline;
}

sub ReportOnComments {
    my $filename = $_[0];

    if ($cxxCommentsFound > 1 && !$gCxxComments) {
	&PrintCurrentDirectory;
	print "$cxx_comments_found C++ comments found in $filename\n";
    }
}

# Add to known actions
$KnownActions[$#KnownActions+1] = "Comments";

# Describe the constructions that we wish to avoid
$gAvoidConstructionsReasons{"//"} = "\
// is not a valid C comment, even though some C compilers accept it.
If this is in a header file used ONLY in C++ programs, include the comment
   /* style:c++ header */
near the top of the file.  Even better, set the Emacs style with
/* -*- Mode: C++; c-basic-offset:4 */";

sub DescribeComments {
    print STDOUT 
"\tcomments - Process comments for special information.  The specific tests
             are controlled by other checks\n";
}

# -----------------------------------------------------------------------------
# These routines read comments in different languages
sub ReadComment {
    my ($FD,$curline,$language) = @_;

    # Handle all languages that have defined a Read<Lanugage>Comment
    my $language = $gLanguage;
    $language =~ s/\+/X/g;
    my $fcn = "Read" . $language . "Comment";
    if (defined(&$fcn)) {
	($comment,$curline) = &$fcn( $FD, $curline );
    }
    else {
	# Assume C
	($comment,$curline) = &ReadCComment( $FD, $curline );
    }

    return ($comment, $curline);
}

# (comment,remainder) = ReadCComment( FD, inputline )
# removes comments from a line and returns the line.  Read more if necessary
sub ReadCComment {
    my ($FD,$curline) = @_;
    my $comment_line = "";
    while ($curline =~ /\/\*/) {
	print "Curline = $curline\n" if $gDebug;
	if ($curline =~ /(\/\*.*?\*\/)/s) {
	    $comment_line = $1;
	    $curline =~ s/\/\*.*?\*\///s;
	    print "Removed comment, now is $curline\n" if $gDebug;
	    # Keep looking for comments incase the line has multiple 
	    # comments
	}
	else {
	    # Keep collecting until we find the end of the comment
	    if (eof($FD)) {
		print STDOUT "Unterminated comment found$errsrc!\n";
		my $line = $curline;
		if ($line =~ /(.*)\n/) { $line = "$1"; }
		print STDOUT "Comment begins with $line\n";
		return ($comment_line,$curline);
	    }
	    $curline .= <$FD>;
	}
    }
    return ($comment_line,$curline);
}

# This reads a C++ comment (//); if it doesn't find one, 
# it tries to read a C comment (/* ... */)
sub ReadCXXComment {
    my ($FD,$curline) = @_;
    my $comment_line = "";

    if ($curline =~ /(.*)\/\/(.*)/) {
	$curline      = $1;
	$comment_line = $2;
    }
    elsif ($curline =~ /\/\*/) {
	($comment_line, $curline) = &ReadCComment($FD,$curline);
    }
    return ($comment_line,$curline);
}

# This handles all Fortran comments, both Fortran 77 and 90
sub ReadFortranComment {
    my ($FD,$curline) = @_;
    my $comment_line = "";

    if ($curline =~ /^[Cc\*!](.*)/) {
	$curline = "";
	$comment_line = $1;
    }
    elsif ($curline =~ /^([^\"\']*)!(.*)/) {
	# Avoids ! in a quoted string.  Does not handle
	# ! after a string (FIXME)
	$curline = $1;
	$comment_line = $2;
    }

    return ($comment_line,$curline);
}

sub ReadPerlComment {
    my ($FD,$curline) = @_;
    return &ReadGeneralCommentToEndOfLine( $FD, $curline, "#" );
}

sub ReadShellComment {
    my ($FD,$curline) = @_;
    return &ReadGeneralCommentToEndOfLine( $FD, $curline, "#" );
}

sub ReadPythonComment {
    my ($FD,$curline) = @_;
    return &ReadGeneralCommentToEndOfLine( $FD, $curline, "#" );
}

# This handles all Script and similar comments
# Handles simple string formats:
#   " xxxx " # comment 
#   '  xxx ' # comment
sub ReadGeneralCommentToEndOfLine {
    my ($FD,$curline,$commentChar) = @_;
    my $comment_line = "";

    # Handle strings
    $inline  = $curline;
    $curline = "";
    while ($inline =~ /([^\"\'$commentChar]*)([\"\'])(.*)/) {
	$curline   .= $1;
	$quotechar = $2;
	$inline    = $3;
	# Find the matching quote
	if ($inline =~ /([^$quotechar]*)$quotechar(.*)/) {
	    $curline .= $quotechar . $1 . $quotechar;
	    $inline = $2;
	}
	else {
	    # Incomplete quote
	    # FIXME: We should probably read the rest of the string
	    # by reading from input
	    $curline .= $inline;
	    $inline   = "";
	    last;
	}
    }
    
    # At this point, $curline contains a non-comment and $inline, if 
    # non-empty, may have a comment.
	
    if ($inline =~ /^([^\"\']*)$commentChar(.*)/) {
	# Avoids comment char in a quoted string.  Does not handle
	# $commentChar after a string (FIXME)
	$curline      .= $1;
	$comment_line  = $2;
    }

    return ($comment_line,$curline);
}

1;


